#!/usr/bin/lua

function save(programs)
	local h=io.open("programs.dat","w")
	for description,lines in pairs(programs) do
		h:write("### "..description.."\n")
		for _,line in ipairs(lines) do
			h:write(line.."\n")
		end
	end
	h:close()
end

function addr(n)
	return string.rep("0",(3-#(n.."")))..n
end

function pad(str,len)
	return str..string.rep(" ",len-#str)
end

function interpret(line)
	local tree={
		[43]={
			[21]="√x",
			[33]="GTO ",
			[24]="FRAC",
			[25]="INTG",
			[34]="x≤y",
			[35]="x=0",
		},
		[16]="CHS",
		[21]="yˣ",
		[31]="R/S",
		[33]="R⬇",
		[34]="x≷y", -- ⇄ ⇌
		[36]="ENTER",
		[44]=function (line,location)
			local op,reg=string.match(string.sub(line,location),"(%d%d),(%d)")
			if reg then
				return "STO "..interpret(op).." "..reg
			else
				return "STO "..string.sub(line,location)
			end
		end,
		[45]="RCL ",
		[48]=".",
		[10]="/", -- ÷
		[20]="×",
		[30]="-",
		[40]="+",
		[0]="0",
		[1]="1",
		[2]="2",
		[3]="3",
		[4]="4",
		[5]="5",
		[6]="6",
		[7]="7",
		[8]="8",
		[9]="9",
	}
	for n,location in string.gmatch(line,"(%d+)()") do
		tree=tree[n/1]
		if not tree then break end
		if type(tree)=="string" then return tree..string.sub(line,location+1) end
		if type(tree)=="function" then return tree(line,location+1) end
	end
	return "[UNKNOWN COMMAND]"
end

function load()
	local programs={}
	local current
	for line in io.lines("programs.dat") do
		local hashes,description=string.match(line,"(### )(.+)")
		if hashes then
			current={}
			programs[description]=current
		else
			table.insert(current,line)
		end
	end
	return programs
end

function menu(items)
	local opt
	repeat
		for i,v in pairs(items) do
			print(i..". "..v[1])
		end
		io.write("Select an option: ") opt=tonumber(io.read())
	until opt and opt>=1 and opt<=#items and math.floor(opt)==opt
	return items[opt][2]()
end

function mainmenu()
	print("Voyager Manager v0.1 for the HP 12c")
	print("Copyright 2024 Tritonio, released under GNU/GPLv3")
	menu{
		{"Add/overwrite program",function ()
			io.write("Program description: ") local description=io.read()
			if description=="" then print("Canceled.") mainmenu() end
			print("Type the program in numerical form, separated by commas, e.g. 241,43,33,054")
			print("The last command must be a GTO 000. End your entry with an empty line.")
			local lines={}
			local baseaddress
			local lastaddress
			local lastisgoto=false
			while true do
				local line=io.read()
				if line=="" then
					if not lastisgoto then print("Last program command should be GTO 000. Entry canceled.") return end
					break
				end
				lastisgoto=false
				local address,line=string.match(line,"^(%d+),([%d,%s]+)$")
				if address==nil then print("Invalid program line. Entry canceled.") return end
				address=tonumber(address)
				if address==nil then print("Invalid address. Entry canceled.") return end 
				if not baseaddress then baseaddress=address-1 end
				address=address-baseaddress
				if address<=0 then print("Invalid address. Entry canceled.") return end
				if not lastaddress then lastaddress=address-1 end
				if lastaddress~=address-1 then print("Non sequential address. Entry canceled.") return end
				lastaddress=address
				line=string.gsub(line,"%s+",",")
				line=string.gsub(line,",+",",")
				local g43,g33,gotoaddress=string.match(line,"(%d%d),(%d%d),(%d%d%d?)")
				if g43=="43" and g33=="33" then
					if tonumber(gotoaddress)~=0 then
						if tonumber(gotoaddress)==nil then print("Invalid GTO address. Entry canceled.") return end
						line="43,33,"..addr(gotoaddress-baseaddress)
					else
						lastisgoto=true
					end
				end
				lines[address]=line
			end
			programs[description]=lines
			print("Program saved.")
			save(programs)
		end},
		{"Compile multi-program",function ()
			print("Available programs:")
			local sorted={}
			for description,lines in pairs(programs) do
				if not string.match(description,"%[HIDDEN%]") then
					table.insert(sorted,{description,lines})
				end
			end
			table.sort(sorted,function (a,b) return a[1]<b[1] end)
			for i,v in ipairs(sorted) do
				print(i..". "..v[1])
			end
			io.write("Write your selections in one line, separated by spaces: ") local selections=io.read()
			local selectionstable={}
			for selection in string.gmatch(selections,"%d+") do
				local item=sorted[selection/1]
				if not item then print("Invalid selection.") return end
				table.insert(selectionstable,item[1]) 
			end
			if #selectionstable==0 then print("No selections were made.") return end
			io.write("You selected "..#selectionstable.." programs. Type the number of spare slots you want: ") local spareslots=io.read()
			if spareslots=="" then spareslots=0 end
			local address=1
			local output=""
			local function printwithaddress(line)
				local line=pad(addr(address)..","..line,15)..interpret(line)
				output=output..line.."\n"
				print(line)
				address=address+1
			end
			local hoppoint=#selectionstable+spareslots+1
			for _,selection in ipairs(selectionstable) do
				printwithaddress("43,33,"..addr(hoppoint))
				hoppoint=hoppoint+#programs[selection]
			end
			for i=1,spareslots do
				printwithaddress("43,33,"..addr(hoppoint))
			end
			hoppoint=#selectionstable+spareslots+1
			for _,selection in ipairs(selectionstable) do
				for _,line in ipairs(programs[selection]) do
					local g43,g33,gotoaddress=string.match(line,"(%d%d),(%d%d),(%d%d%d?)")
					if g43=="43" and g33=="33" and tonumber(gotoaddress)~=0 then
						line="43,33,"..addr(gotoaddress+hoppoint-1)
					end
					printwithaddress(line)
				end
				hoppoint=hoppoint+#programs[selection]
			end
			local h=io.open("multi-program.txt","w")
			h:write(output)
			h:close()
			print("Also saved in multi-program.txt.")
		end},
		{"Hide program(s)",function ()
			print("Available programs:")
			local sorted={}
			for description,lines in pairs(programs) do
				if not string.match(description,"%[HIDDEN%]") then
					table.insert(sorted,{description,lines})
				end
			end
			table.sort(sorted,function (a,b) return a[1]<b[1] end)
			for i,v in ipairs(sorted) do
				print(i..". "..v[1])
			end
			io.write("Write your selections in one line, separated by spaces: ") local selections=io.read()
			local selectionstable={}
			for selection in string.gmatch(selections,"%d+") do
				local item=sorted[selection/1]
				if not item then print("Invalid selection.") return end
				table.insert(selectionstable,item[1]) 
			end
			if #selectionstable==0 then print("No selections were made.") return end
			io.write("You selected "..#selectionstable..' programs. Type "hide" to hide them: ') local confirmation=io.read()
			if string.lower(confirmation)=="hide" then
				for _,selection in ipairs(selectionstable) do
					programs[selection],programs["[HIDDEN]"..selection]=nil,programs[selection]
				end
				print("Deleted.")
				save(programs)
			else
				print("Canceled.") 
			end
		end},
		{"Exit",os.exit}
	}
end


programs=load()

while true do
	mainmenu()
end